"""
Module constants for JACAMAR.
"""

import flint

CONST_E = flint.arb.const_e()
PI = flint.arb.pi()
CONST_LOG2 = flint.arb.const_log2()
CONST_EULER = flint.arb.const_euler()
PACKING_BOUND = 2**16
MATRIX_SWITCH = 249  # TODO: determine this number during build via tests
